
-- File Mailer Engine

-- Note: the File Mailer was originally only designed for sending
-- attachments. The new 'soliciting' feature was added later. The mods
-- use 'attach' in the DB for storing 'soliciting state' info.

local fmt=string.format
local tinsert=table.insert
local quote=luasql.quotestr
local rio=ba.openio"root"
local sqlutil=require"sqlutil"
local mailutil=require"mailutil"
require"socket.mail"

local engineActive=true -- Can be stopped by user
local mailEngineCo -- mailEngine coroutine
local timer      -- BA timer obj
local startjob   -- function
local startEngine -- function

-- The following is set by function loadSettings
local timerTickVal -- The time between sending emails
local domainname  -- Used when creating download links
local fromaddr -- The default email from addr
local mail  -- A socket.mail created object
local timerTickValAdjusted -- == timerTickVal: Increased on error

local trun -- Thread run
do
   local t
   trun=function(func)
      local thread = t and t["thread"]
      if not thread then
         thread=ba.thread.create()
         t=setmetatable({thread=thread},{__mode="v"})
      end
      thread:run(func)
   end
end


-- Assemble error response msg
local function eResp(...)
   local emsg=table.concat({...}, ": ")
   trace("Filemailer:",emsg)
   mailutil.seterr(emsg)
   return nil,emsg
end

-- Generic response checker. Returns first arg on success.
local function checkResp1(...)
   local args={...}
   if not args[1] then return eResp(...) end
   return args[1]
end


local function openDb() 
   local env,conn=bd.openDb"bd"
   conn:setbusytimeout(10000)
   return env,conn
end


-- Open db, execute sql, and close db
local function dbexec(sql)
   local env,conn=openDb()
   conn:setautocommit("IMMEDIATE",30000)
   local ok,err,err2=conn:execute(sql)
   if ok then
      conn:commit()
   else
      conn:rollback()
   end
   bd.closeDb(env,conn)
   if not ok then return eResp(err,err2) end
   return ok
end

-- Load settings from database
local function loadSettings()
   local env,conn=openDb()
   local email,server,port,useauth,uname,password,consec=sqlutil.find(
      conn,"email,server,port,useauth,uname,password,consec FROM mailsmtp")
   local cfg={server=server or "", port=port and tonumber(port) or 25}
   if useauth=="1" then
      cfg.user=uname
      cfg.password=bd.aesdecode(password)
      if consec == "tls" then
         cfg.shark=bd.getDefaultSharkClient()
      elseif consec == "starttls" then
         cfg.starttls=true
         cfg.shark=bd.getDefaultSharkClient()
      end
   end
   mail=socket.mail(cfg)
   fromaddr=email
   timerTickVal=sqlutil.find(conn,"value FROM mailsettings WHERE key='ticktime'")
   timerTickVal = timerTickVal and tonumber(timerTickVal) or 60000
   timerTickVal = timerTickVal > 500 and timerTickVal or 500
   timerTickValAdjusted=timerTickVal
   domainname=sqlutil.find(conn,"value FROM mailsettings WHERE key='domainname'")
   domainname=domainname or "domainname"
   bd.closeDb(env,conn)
end

local function createAddr(firstname,lastname,email)
   return (#firstname > 0 or #lastname > 0) and fmt("%s %s <%s>",firstname,lastname,email) or email
end

local function expandMacros(cfg,id,links,txtbody,htmlbody,firstname,lastname)
   local action = {ID=function() return id or "NO ID" end}
   if links then
      action.LINK=function() return table.concat(links,"\n") end
   end
   if firstname then
      action.FIRSTNAME=function() return #firstname > 0 and firstname or lastname end
      action.LASTTNAME=function() return #lastname > 0 and lastname or firstname end
      action.NAME=function() return fmt("%s %s", firstname, lastname) end
   end
   cfg.txtbody=mailutil.expandmacros(txtbody, action)
   if htmlbody then
      action.LINK=function() 
         local t={}
         for _,link in ipairs(links) do
            tinsert(t,fmt('<a href="%s">%s</a><br/>',link,link))
         end
         return table.concat(t)
      end
      cfg.htmlbody=mailutil.expandmacros(htmlbody, action)
   end
end


local function getErrCode(err)
   if err then
      err=err:match"^%s*(%d+)"
      if err then return tonumber(err) end
   end
end


local function is5xxErr(err)
   err=getErrCode(err)
   if err then
      err=err/100
      if err >= 5 and err < 6 then return true end
   end
end


local function sendEmail(cfg,setAttach)
   while engineActive do
      setAttach(cfg)

      local ok,err=mail:send(cfg)

      collectgarbage("step",2)
      if ok then
         timerTickValAdjusted=timerTickVal
         return
      end
      if timerTickValAdjusted < 1000*60*60*12 then
         timerTickValAdjusted=timerTickValAdjusted*2
      end
      if timerTickValAdjusted < 60000 then
         timerTickValAdjusted=60000
      end
      if type(cfg.to) == "string" then
         eResp(fmt("Addr: %s, Error: %s",cfg.to,err))
      else
         eResp(err)
      end
      if is5xxErr(err) then return end
      coroutine.yield(true) -- Sleep
   end
   -- Abort
   mailutil.setjob()
   coroutine.yield(false)
end



-- Runs as a BA thread
local function mailerThread()
   local ok, resume
   if engineActive then
      ok, resume=coroutine.resume(mailEngineCo)
   end
   if engineActive then
      if ok then
         if resume then
            timer:set(timerTickValAdjusted,true)
            return
         end
      elseif resume then -- resume=err
         eResp("mailerThread err:",resume)
      end
   end
   mailEngineCo=nil
   if not engineActive then
      startEngine() -- Restart
   end
end


local function mailerTick()
   trun(mailerThread)
end


-- The mailEngine coroutine runs in the context of mailerThread
local function mailEngine()
   local gc=startjob -- Prevent GC
   while true do
      local ok,err
      local files -- set to list of files if not attach
      local mj={} -- Mailjob table
      sqlutil.findt(openDb,"key,fromaddr,hash,subject,txtbody,htmlbody,riop,tokey,listkey,groupkey,sendcnt,attach,secure,password FROM mailjob where active=1 limit 1",mj)
      if not mj.key then break end -- No more jobs
      local sendcnt=tonumber(mj.sendcnt)
      local passwdreq=mj.password == "1"
      local htmlbody = #mj.htmlbody>0 and mj.htmlbody or nil
      local txtbody = (not htmlbody or #mj.txtbody>0) and mj.txtbody or nil
      local cfg
      if mj.attach == "solicit" then
         cfg={
            from=mj.fromaddr,
            subject=mj.subject,
         }
      else
         cfg={
            from=fromaddr,
            replyto=mj.fromaddr,
            subject=mj.subject,
         }
      end

      if not htmlbody and txtbody:find("multipart/related") then
         local m = require'MIME/EmailMessage'(txtbody)
         if m then
            txtbody,htmlbody = m.txtbody,m.htmlbody
            cfg.htmlimg=m.htmlimg
            cfg.charset=m.charset
         end
      end

      local listkey=ba.json.decode(mj.listkey)
      local secure = mj.secure == "1"
      if mj.riop then
         local getlinks -- Function
         local setAttach -- Function
         if mj.attach then
            local io
            io,err=checkResp1(ba.mkio(rio,mj.riop),mj.riop)
            if not io then break end
            local attach=ba.json.decode(mj.attach)
            setAttach=function(cfg)
               local t={}
               cfg.attach=t
               for _,v in ipairs(attach) do
                  local fp
                  fp,err=checkResp1(io:open(v),mj.riop,v)
                  if not fp then break end
                  tinsert(t, {source=fp,name=v})
               end
            end
            getlinks = function() end -- Return NIL
         else
            local filekeys={}
            local function assemble(cur)
               local key=cur:fetch()
               while key do
                  tinsert(filekeys,key)
                  key=cur:fetch()
               end 
            end
            setAttach = function() end
            getlinks = function(contacthash)
               links={}
               for _,k in ipairs(filekeys) do
                  assert(not contacthash or #mj.hash==2)
                  local h=fmt("http%s://%s/rtl/e/%s%s%s%x",
                              secure and "s" or "",
                              domainname,
                              contacthash and string.char(#contacthash+45) or "",
                              contacthash or "",
                              mj.hash,
                              k)
                  tinsert(links,h)
               end
               return links
            end
            ok,err=sqlutil.select(openDb,"key,name FROM mailjobfile WHERE mailjobkey="..mj.key,assemble)
         end
         if not err then
            if mj.tokey then -- Send to individual(s)
               local contacthash,id -- Defaults to nil
               mailutil.setjob(mj.key)
               local to=ba.json.decode(mj.tokey)
               local bcc=ba.json.decode(mj.groupkey)
               cfg.to = to
               cfg.cc = #listkey > 0 and listkey or nil
               cfg.bcc = #bcc > 0 and bcc or nil
               if #to == 1 and (#listkey + #bcc) == 0 then
                  local name,email=to[1]:match"([^<]*)<([^>]*)>.*"
                  if not name then email=to[1] end
                  local sql=fmt("%s%s","id,hash FROM mailcontact WHERE email=",quote(email))
                  id,contacthash,e=sqlutil.find(openDb,sql)
               end
               expandMacros(cfg,id,getlinks(contacthash),txtbody,htmlbody)
               sendEmail(cfg,setAttach)
            else -- Send to list(s)
               while true do
                  local sql=fmt("%s%s)%s LIMIT %u,1",
                                "DISTINCT email,firstname,lastname,id,hash FROM mailcontact WHERE (listkey=",
                                table.concat(listkey, " or listkey="),
                                passwdreq and " AND password IS NOT NULL" or "",
                                sendcnt)
                  local email,firstname,lastname,id,contacthash=sqlutil.find(openDb,sql)
                  if not firstname then break end
                  mailutil.setjob(mj.key,sendcnt)
                  cfg.to = createAddr(firstname,lastname,email)
                  local links
                  expandMacros(cfg,id,getlinks(contacthash),txtbody,htmlbody,firstname,lastname)
                  sendEmail(cfg,setAttach)
                  sendcnt=sendcnt+1
                  dbexec(fmt("UPDATE mailjob SET sendcnt=%u WHERE key=%s",sendcnt,mj.key))
                  coroutine.yield(true) -- Sleep
               end
            end
         end
      else -- not riop: Send password confirmation emails
         local function setAttach() end
         local action =  mj.attach and "" or "?a=pwd"
         while true do
            local sql
            if #listkey > 0 then
               if mj.attach then -- solicit
                  sql=fmt("%s%s)%s LIMIT %u,1",
                          "DISTINCT email,firstname,lastname,id,hash FROM mailcontact WHERE (listkey=",
                          table.concat(listkey, " or listkey="),
                          passwdreq and " AND password IS NOT NULL" or "",
                          sendcnt)
               else
                  sql=fmt("%s AND (listkey=%s) LIMIT %u,1",
                          "DISTINCT email,firstname,lastname,id,hash FROM mailcontact WHERE password IS NULL",
                          table.concat(listkey, " or listkey="),sendcnt)
               end
            else
               sql="DISTINCT email,firstname,lastname,id,hash FROM mailcontact WHERE password IS NULL LIMIT %u,1"
            end
            local email,firstname,lastname,id,contacthash=sqlutil.find(openDb,sql)
            if not email then
               if firstname then -- err
                  trace(sql,firstname)
                  collectgarbage()
                  ba.sleep(1000)
                  email,firstname,lastname,id,contacthash=sqlutil.find(openDb,sql)
               end
               if not email then break end
            end
            mailutil.setjob(mj.key,sendcnt)
            local links={fmt("http%s://%s/rtl/e/%s%s%s",
                             secure and "s" or "",
                             domainname,
                             string.char(#contacthash+45),
                             contacthash,
                             action)}
            expandMacros(cfg,id,links,txtbody,htmlbody,firstname,lastname)
            cfg.to = createAddr(firstname,lastname,email)
            sendEmail(cfg,setAttach)
            sendcnt=sendcnt+1
            dbexec(fmt("UPDATE mailjob SET sendcnt=%u WHERE key=%s",sendcnt,mj.key))
            coroutine.yield(true) -- Sleep
         end
      end
      dbexec("UPDATE mailjob SET active=0 WHERE key="..mj.key)
      collectgarbage()
      coroutine.yield(true) -- Sleep
   end
   mailutil.setjob()
   coroutine.yield(false)
end


startEngine=function()
   engineActive=true
   loadSettings()
   timer=ba.timer(mailerTick)
   mailEngineCo=coroutine.create(mailEngine)
   trun(mailerThread)
end



--[[
table e: {
  io    = BA I/O where files are stored
  path  = path to directory within "io"
  files = list of files

  individual = table or nil if list set
  list = the list(s) or nil if individual set
  settings = table

  body = Email body
  subject = the subject or nil
  defsub  = Default (system generated) subject, used if subject=nil
}

table settings: {
attach = true or nil: send files as attachement or links
ssl = true or nil: Secure link required
pwd = true or nil: Password required for link
exptime = secs since January 1 1970 or nil if link never expires

table individual: {
  to = list of email addresses
  cc = list of email addresses
  bcc = list of email addresses
}
--]]
startjob=function(user,e,stop)
   local function run()
      if mailEngineCo then -- Engine created
         if not user then -- If updating settings
            if stop then
               engineActive=false
               timerTickValAdjusted=1
               timer:reset(1)
               -- Will restart in mailerThread
            else
               loadSettings()
               timer:reset(timerTickVal)
            end
         end
      else
         startEngine()
      end
   end
   local function dos2unix(path)
      if not path then return "/" end
      if path:find"^(%w)(:)" then
         path=path:gsub("^(%w)(:)", "%1",1)
      end
      return path:gsub("\\", "/")
   end
   if user and e then
      local fromaddr=bd.db.email(user) or user
      local hash=fmt("%x",ba.clock()):sub(-2,-1)
      local hash=fmt("%08x",ba.clock()):sub(1,8)
      if not fromaddr then return eResp"No sender address" end
      local subject=e.subject or e.defsub
      local txtbody=e.bodyishtml and "" or e.body
      local htmlbody=e.bodyishtml and e.body or ""
      local riop = e.path and dos2unix(e.io:realpath"")..e.path
      if riop and not ba.mkio(ba.openio"root",riop) then return eResp("Invalid IO or path") end
      local settings=e.settings
      local attach=settings.attach and quote(ba.json.encode(e.files)) or "NULL"
      local secure=settings.ssl and 1 or 0
      local password=settings.pwd and 1 or 0
      local sql
      local batchkey = settings.batchkey and tonumber(settings.batchkey) and settings.batchkey or "NULL"
      if not riop then  -- Send soliciting emails or password confirmation emails
         sql=fmt("%s VALUES(%s,%s,%s,%s,%s,%s,1,0,%s,%u,%u)",
                 "INSERT INTO mailjob (bduser,fromaddr,subject,txtbody,htmlbody,listkey,active,sendcnt,attach,secure,date)",
                 quote(user),
                 quote(fromaddr),
                 quote(subject),
                 quote(txtbody),
                 quote(htmlbody),
                 quote(ba.json.encode(e.list)),
                 e.solicit and "'solicit'" or "NULL",
                 secure,
                 os.time())
      elseif e.individual then
         local addr=e.individual
         if #addr.to == 0 then
            return nil,'At least one "to" field is required'
         end
         hash = #addr.to+#addr.cc+#addr.bcc == 1 and hash:sub(-2,-1) or hash
         sql=fmt("%s VALUES(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,1,0,%s,%u,%u,%u,%u,%s)",
                 "INSERT INTO mailjob (bduser,fromaddr,hash,subject,txtbody,htmlbody,riop,tokey,listkey,groupkey,active,sendcnt,attach,secure,password,expires,date,batchkey)",
                 quote(user),
                 quote(fromaddr),
                 quote(hash),
                 quote(subject),
                 quote(txtbody),
                 quote(htmlbody),
                 quote(riop),
                 quote(ba.json.encode(addr.to)),
                 quote(ba.json.encode(addr.cc)),
                 quote(ba.json.encode(addr.bcc)),
                 attach,
                 secure,
                 password,
                 settings.exptime or 0,
                 os.time(),
                 batchkey)
      else
         hash=hash:sub(-2,-1)
         sql=fmt("%s VALUES(%s,%s,%s,%s,%s,%s,%s,%s,%s,1,0,%s,%u,%u,%u,%u,%s)",
                 "INSERT INTO mailjob (bduser,fromaddr,hash,subject,txtbody,htmlbody,riop,listkey,groupkey,active,sendcnt,attach,secure,password,expires,date,batchkey)",
                 quote(user),
                 quote(fromaddr),
                 quote(hash),
                 quote(subject),
                 quote(txtbody),
                 quote(htmlbody),
                 quote(riop),
                 quote(ba.json.encode(e.list)),
                 quote(ba.json.encode{}),
                 attach,
                 secure,
                 password,
                 settings.exptime or 0,
                 os.time(),
                 batchkey)
      end
      local env,conn=openDb()
      if not settings.attach and #mailutil.settings(conn,"domainname") == 0 then
         bd.closeDb(env,conn)
         return false,'<a target="_blank" href="../manage/settings.lsp">Domain name</a> not set.'
      end

      conn:setautocommit("IMMEDIATE",30000)
      ok,err,err2=conn:execute(sql)
      if ok and not settings.attach and e.files then
         local mailjobkey=conn:lastid()
         for _,file in ipairs(e.files) do
            sql=fmt("%s VALUES(%u,%s)",
                    "INSERT INTO mailjobfile (mailjobkey,name)",
                    mailjobkey,quote(file))
            ok,err,err2=conn:execute(sql)
            if not ok then break end
         end
      end
      if ok then
         conn:commit()
      else
         ok,err=eResp(err,err2)
         conn:rollback()
      end
      bd.closeDb(env,conn)
      if ok then run() end
      return ok,err
   end
   run()
   return true
end

local t={
   startjob=startjob,
   sendmail = function(cfg)
      if not mail then loadSettings() end
      if not cfg.to then error("to address missing",2) end
      cfg.from = cfg.from or fromaddr
      cfg.subject = cfg.subject or ""
      return mail:send(cfg)
   end
}

return setmetatable(t, { __mode = 'v' })
